import os



##############################################################################################################
#
# NOTE !!!
#
# THIS FILE MAY BE AUTOGENERATED.
# ONLY MODIFY THE ORIGINAL FILE IN THE /sharedData/ FOLDER!
# OTHER INSTANCES OF THIS FILE ARE COPIES!
#
# NOTE !
#
##############################################################################################################


class Identifier:
    """
    a token by which objects of different types can be compared.
    All objects that are read from the input have an identifier.
    All objects generated during the output have a preliminary identifier, because the output will be interpreted by the Execution Environment.
    (preliminary identifiers can be used like any other).
    An identifier may optionally also have a name.
    Identifiers provided by the server will have this field set if the Identifier refers to an object with a name.
    This fields is optional and not used for identification purposes.
    """
    def __init__(self, identifier, type, preliminary=False, name=None):
        self.id = identifier
        self.type = type
        self.preliminary = preliminary
        self.name = name
    def __str__(self):
        return "%s Identifier %s%s" % (self.type, self.id, " (preliminary)" if self.preliminry else "")
    def __hash__(self):
        """
        override the default hashing, to make equality comparisons possible
        """
        res = hash(self.type)
        res *= 17
        res += hash(self.id)
        res *= 17
        res += hash(self.preliminary)
        return res
    def __eq__(self, other):
        """
        Override the default Equals behavior
        """
        return (self.type == other.type) and (self.id == other.id) and (self.preliminary == other.preliminary)
    def __ne__(self, other):
        """
        Define a non-equality test
        """
        return not self.__eq__(other)
    def to_json(self):
        """
        gives a JSON dictionary representation of this identifier.
        Counterpart to parse_identifier().
        """
        res = { 'type' : self.type, 'id' : self.id }
        if self.preliminary:
            res['preliminary'] = True
        if self.name:
            res['name'] = self.name
        return res


def parse_identifier(dictionary):
        """
        creates an Identifier from a JSON dictionary structure
        Counterpart to Identifier.to_json().
        """
        id = dictionary['id']
        type = dictionary['type']
        preliminary = dictionary.get('preliminary', False)
        name = dictionary.get('name', None)
        return Identifier(id, type, preliminary=preliminary, name=name)


_preliminary_dentifier_counter = 0
def create_preliminary_identifier(type):
    """
    creates a preliminary Identifier, for use by output objects.
    """
    global _preliminary_dentifier_counter
    res = Identifier(_preliminary_dentifier_counter, type, preliminary=True)
    _preliminary_dentifier_counter += 1
    return res


class FileObject:
    """
    represents a file that was given as an input argument, along with information about it.
    """
    def __init__(self, identifier, file_name, creation_step, creation_index, usage_index=None):
        self.identifier = identifier
        self.file_name = file_name
        self.creation_step = creation_step
        self.creation_index = creation_index
        self.usage_index = usage_index # this is None for newly created output files, and an index for input files
    def __str__(self):
        return self.identifier
    def _change_file_name_for_use_as_input(self, index):
        """
        replaces the name of the file with a name that is used for input files.
        This method should be used by the ExecutionEnvironment after copying a file to use it as input for another program execution.
        """
        self.file_name = "in_%d" % index
    def to_json(self):
        """
        gives a JSON dictionary representation of this FileObject that can be parsed as a new FileObject.
        this is a counterpart to parse_file_object().
        """
        res = {}
        res['file'] =  self.file_name
        res['identifier'] = self.identifier.to_json()
        res['creation_step'] = self.creation_step
        res['creation_index'] = self.creation_index
        return res


def parse_file_object(dictionary, usage_index=None):
    """
    creates a FileObject from a JSON dictionary structure
    Counterpart to FileObject.to_json()
    """
    identifier = parse_identifier(dictionary['identifier'])
    file_name = dictionary['file']
    creation_step = dictionary['creation_step']
    creation_index = dictionary['creation_index']
    res = FileObject(identifier, file_name, creation_step, creation_index, usage_index)
    return res


class Tag:
    """
    a Tag consists of a Symbol, defined here as a string, and a number of identifiers.
    It also has an identifier of its own.
    """
    def __init__(self, own_identifier, symbol, *arguments):
        if not isinstance(own_identifier, Identifier):
            raise ValueError("a Tag must get an Identifier as its identifier")
        if not isinstance(symbol, str):
            raise ValueError("the Symbol of a Tag must be given as a string")
        arguments = [self._get_identifier_of_argument(a) for a in arguments]
        self.identifier = own_identifier
        self.symbol = symbol
        self.arguments = arguments
    def __str__(self):
        return "Tag %s: symbol=%s, arguments=(%s)" % (self.identifier, self.symbol, ', '.join(['%s' % a for a in identifiers]))
    def to_json(self):
        """
        gives a JSON dictionary representation of this Tag.
        Counterpart to parse_tag().
        """
        res = {'identifier' : self.identifier.to_json(), 'symbol' : self.symbol, 'args' : [self._get_identifier_of_argument(a).to_json() for a in self.arguments]}
        return res
    def _get_identifier_of_argument(self, arg):
        if isinstance(arg, Identifier):
            return arg
        if hasattr(arg, 'identifier') and isinstance(arg.identifier, Identifier):
            return arg.identifier
        raise ValueError("each argument of a Tag must either be an Identifier or have an identifier field of that type")


def parse_tag(dictionary):
    """
    creates a Tag from a JSON dictionary structure
    Counterpart to Tag.to_json().
    """
    identifier = parse_identifier(dictionary['identifier'])
    symbol = dictionary['symbol']
    args = [parse_identifier(a) for a in dictionary['args']]
    return Tag(identifier, symbol, *args)


class EventRequest:
    """
    represents a request of this program to the Execution Environment.
    """
    def __init__(self, request_type):
        self.request_type = request_type
    def to_json(self):
        """
        gives a JSON dictionary representation of this EventRequest.
        """
        res = {
            'event_type' : self.request_type,
            'args' : self._args_to_json(),
        }
        return res


class ProgramExecutionRequest(EventRequest):
    """
    a request to execute a Program.
    This expects a program as its first argument, and files to use as parameters as the rest.
    """
    def __init__(self):
        super().__init__('execute_program')
        self.program_identifier = None
        self.argument_identifiers = []
    def program(self, program_identifier):
        if not isinstance(program_identifier, str) and not isinstance(program_identifier, int):
            error_message = """The program may be identified as a String displaying the name of the program (in which case the latest version is picked),
                as a String of <name>#<version> (which identifies the version directly),
                or as an integer that is the program's ID (which is unambiguous and includes the version)"""
            raise ValueError(error_message)
        self.program_identifier = program_identifier
        return self
    def arguments(self, argument_identifiers):
        for arg in argument_identifiers:
            if not isinstance(arg, str) and not isinstance(arg, Identifier):
                raise ValueError("each argument of a program execution request must be a string or an Identifier")
        self.argument_identifiers = argument_identifiers
        return self
    def _args_to_json(self):
        res = {
            'program_identifier' : self.program_identifier,
            'argument_identifiers' : [a.to_json() for a in self.argument_identifiers],
        }
        return res


class DisplayMessageRequest(EventRequest):
    """
    a request to display a message to the user in an interactive website.
    The message generated here will be put in a HTML form and displayed.
    """
    def __init__(self):
        super().__init__('display_message')
        self.message_text = None
        self.add_feedback_buttons = False
    def text(self, message_text):
        if not isinstance(message_text, str):
            raise ValueError("The message to display must be a string")
        self.message_text = message_text
        return self
    def feedback_buttons(self, add=True):
        self.add_feedback_buttons = add
        return self
    def _args_to_json(self):
        res = {
            'message_text' : self.message_text,
            'add_feedback_buttons' : self.add_feedback_buttons
        }
        return res





import json
import os
import sys
import traceback

from .basics import Identifier, parse_identifier, create_preliminary_identifier, FileObject, parse_file_object, Tag, parse_tag, EventRequest, ProgramExecutionRequest, DisplayMessageRequest


##############################################################################################################
# constants
##############################################################################################################


PATH_PREFIX = '/collab'
INPUT_FILE_LOCATION = os.path.join(PATH_PREFIX, 'in.txt')
OUTPUT_FILE_LOCATION = os.path.join(PATH_PREFIX, 'out.txt')

_SOURCE_OF_TRUTH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'TRUTH.txt')
_truth = None
def _get_truth():
    """
    load constants from a file that is autogenerated from the outside when this module is compiled.
    This is a convenient way to share constants between different applications during development.
    Uses a cache.
    """
    global _truth
    if _truth is None:
        with open(_SOURCE_OF_TRUTH, 'r') as f:
            _truth = json.load(f)
    return _truth

def _get_valid_identifier_types():
    """
    returns a list of strings that are all the valid types of identifiers.
    """
    res = [a[0] for a in _get_truth()['valid_identifier_types']]
    return res

def _get_valid_event_request_names():
    """
    returns a list of strings that are all the valid types of event requests.
    """
    res = [a[0] for a in _get_truth()['valid_event_request_types']]
    return res


##############################################################################################################
# a context manager to take care of logging and error handling
##############################################################################################################


_the_manager = None
class manager:
    """
    Context manager for managing logging and error handling.
    """
    def __init__(self, suppress_exceptions_after_logging_them=False, redirect_stdout_to_log=True):
        global _the_manager
        if _the_manager is None:
            _the_manager = self
        else:
            warning = "only one manager of the collab library should be created!"
            print(warning)
            raise IOException(warning)
        self.suppress_exceptions_after_logging_them = suppress_exceptions_after_logging_them
        self.error_file = os.path.join(PATH_PREFIX, 'error.txt')
        self.log_file = os.path.join(PATH_PREFIX, 'log.txt')
        self.redirect_stdout_to_log = redirect_stdout_to_log
    def __enter__(self):
        self.log_file_stream = open(self.log_file, 'w')
        self.log_file_stream.__enter__()
        if self.redirect_stdout_to_log:
            self.previous_stdout = sys.stdout
            sys.stdout = self.log_file_stream
        return self
    def __exit__(self, etype, value, exception_traceback):
        # stop redirecting stdout to the log file
        if self.redirect_stdout_to_log:
            sys.stdout = self.previous_stdout
        # close the log file
        self.log_file_stream.__exit__()
        # make the output available
        make_output_available()
        # log any uncaught exceptions that may have occurred
        if exception_traceback:
            error_message = "%s\n-----\n%s\n-----\n" % (etype, value,)
            with open(self.error_file, 'w') as f:
                f.write(error_message)
                traceback.print_tb(exception_traceback, limit=None, file=f)
            # allow the exception to continue
            return self.suppress_exceptions_after_logging_them
        else:
            # no error, so nothing needs to be done
            pass
    def log(self, message):
        """log a message to the log file.
        This file is available via the collab website for inspection, but is not available to other programs running in collab."""
        self.log_file_stream.write(message)
        self.log_file_stream.write('\n')


def log(message):
    """uses the context manager for logging"""
    if _the_manager is None:
        raise IOError("the collab library's manager needs to be created in order to use logging.")
    _the_manager.log(message)


##############################################################################################################
# input
##############################################################################################################


_input = None
def _get_input():
    """
    parse the input file. This uses a cache, so it only has to be done once.
    """
    global _input
    if _input is None:
        with open(INPUT_FILE_LOCATION, 'r') as f:
            _input = json.load(f)
    return _input


def get_current_step():
    """
    returns an integer indicating the number of the current step in the execution environment.
    """
    return _get_input()['current_step']


def get_inputs():
    """
    returns a list of inputs that were given to this program.
    """
    inputs = _get_input()['inputs']
    res = []
    for c,a in enumerate(inputs):
        obj = parse_file_object(a, c)
        obj.file = os.path.join(PATH_PREFIX, obj.file_name)
        res.append(obj)
    return res


class ProgramTrigger:
    """
    describes why this program was triggered
    """
    def __init__(self, event_type, reason):
        self.event_type = event_type
        self.reason = reason


def get_trigger():
    """
    returns a ProgramTrigger that describes why this program was called
    """
    trigger = _get_input()['trigger']
    res = ProgramTrigger(trigger['event_type'], trigger['reason'])
    return res


_all_earlier_tags = None
def get_tags():
    """
    returns a list of tags that were created earlier and given to this program as input.
    """
    global _all_earlier_tags
    if _all_earlier_tags is None:
        tags = _get_input()['tags']
        _all_earlier_tags = [parse_tag(a) for a in tags]
    return _all_earlier_tags


##############################################################################################################
# output
##############################################################################################################


_output_counter = 0
_output_files = []
def add_output_file():
    """
    create an output, to make the results of this program available to other programs and to the Execution Environment.
    you can create multiple outputs, and they will be added in the order in which they were created with this function.
    """
    global _output_counter
    global _output_files
    identifier = create_preliminary_identifier('file')
    file_name = "out_%d" % _output_counter
    creation_index = _output_counter
    creation_step = get_current_step()
    res = FileObject(identifier, file_name, creation_step, creation_index)
    res.file = os.path.join(PATH_PREFIX, res.file_name)
    _output_counter += 1
    _output_files.append(res)
    return res


_output_tags = []
def create_tag(symbol, *arguments):
    """
    creates a new Tag using a preexisting Collab Symbol and an arbitrary number of identifiers.
    Note that this Tag is preliminary and will be rejected by the Execution Environment if the Symbol does not actually exist,
    or is not accessible for this program, or if anything else is wrong.
    For any arguments, you may use either an Identifier or any class that has an identifier as a field of the same name,
    which is all of the classes used here.
    """
    global _output_tags
    identifier = create_preliminary_identifier('tag')
    res = Tag(identifier, symbol, *arguments)
    _output_tags.append(res)
    return res


_output_event_requests = []
def _add_event_request(request):
    """
    add a request to the Execution Environment.
    Note that this function works differently depending on the type of the request.
    """
    global _output_event_requests
    request_type = request.request_type
    if not request_type in _get_valid_event_request_names():
        raise ValueError("'%s' is not a valid type of event request" % (request_type,))
    _output_event_requests.append(request)


def request_program_execution(program_identifier, *argument_identifiers):
    """
    a request to execute a Program.
    This expects a program as its first argument, and files to use as parameters as the rest.
    The program may be identified as a String displaying the name of the program (in which case the latest version is picked),
    as a String of <name>#<version> (which identifies the version directly),
    or as an integer that is the program's ID (which is unambiguous and includes the version)
    """
    request = ProgramExecutionRequest().program(program_identifier).arguments(argument_identifiers)
    _add_event_request(request)
    return request


def request_to_display_message(message_text):
    """
    returns a DisplayMessageRequest, which is used for building a message to display to the user,
    with several options that can be added via method chaining.
    """
    request = DisplayMessageRequest().text(message_text)
    _add_event_request(request)
    return request


def make_output_available():
    """
    create the output file and write all previously added messages to the execution environemnt into it.
    This function is called automatically when the context manager closes, but it is possible to call it earlier, too,
    if you are unsure if the program might time out later, so that you can at least output some preliminary results.
    """
    res = {
        'output_files' : [a.to_json() for a in _output_files],
        'output_tags' : [a.to_json() for a in _output_tags],
        'event_requests' : [a.to_json() for a in _output_event_requests],
    }
    with open(OUTPUT_FILE_LOCATION, 'w') as f:
        json.dump(res, f, indent=4)



